from typing import List

import pytest
from mock import AsyncMock

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control.emulation.module_server import (
    ModuleStatusClient,
    helpers,
    models,
)
from opentrons.hardware_control.modules import ModuleAtPort


@pytest.fixture
def mock_callback() -> AsyncMock:
    """Callback mock."""
    return AsyncMock(spec=helpers.NotifyMethod)


@pytest.fixture
def mock_client() -> AsyncMock:
    """Mock client."""
    return AsyncMock(spec=ModuleStatusClient)


@pytest.fixture
def subject(mock_callback: AsyncMock, mock_client: AsyncMock) -> helpers.ModuleListener:
    """Test subject."""
    return helpers.ModuleListener(client=mock_client, notify_method=mock_callback)


@pytest.fixture
def connections() -> List[models.ModuleConnection]:
    """Connection models."""
    return [
        models.ModuleConnection(
            url=f"url{i}", module_type=f"module_type{i}", identifier=f"identifier{i}"
        )
        for i in range(5)
    ]


@pytest.fixture
def modules_at_port() -> List[ModuleAtPort]:
    """Connection models."""
    return [
        ModuleAtPort(
            port=f"url{i}",
            name=f"module_type{i}",
            usb_port=USBPort(name=f"identifier{i}", port_number=i + 1),
        )
        for i in range(5)
    ]


async def test_handle_message_connected_empty(
    subject: helpers.ModuleListener, mock_callback: AsyncMock
) -> None:
    """It should call the call back with the correct modules to add."""
    message = models.Message(status="connected", connections=[])
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with([], [])


async def test_handle_message_connected_one(
    subject: helpers.ModuleListener,
    mock_callback: AsyncMock,
    connections: List[models.ModuleConnection],
    modules_at_port: List[ModuleAtPort],
) -> None:
    """It should call the call back with the correct modules to add."""
    message = models.Message(status="connected", connections=connections[:1])
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with(modules_at_port[:1], [])


async def test_handle_message_connected_many(
    subject: helpers.ModuleListener,
    mock_callback: AsyncMock,
    connections: List[models.ModuleConnection],
    modules_at_port: List[ModuleAtPort],
) -> None:
    """It should call the call back with the correct modules to add."""
    message = models.Message(status="connected", connections=connections)
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with(modules_at_port, [])


async def test_handle_message_disconnected_empty(
    subject: helpers.ModuleListener, mock_callback: AsyncMock
) -> None:
    """It should call the call back with the correct modules to remove."""
    message = models.Message(status="disconnected", connections=[])
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with([], [])


async def test_handle_message_disconnected_one(
    subject: helpers.ModuleListener,
    mock_callback: AsyncMock,
    connections: List[models.ModuleConnection],
    modules_at_port: List[ModuleAtPort],
) -> None:
    """It should call the call back with the correct modules to remove."""
    message = models.Message(status="disconnected", connections=connections[:1])
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with([], modules_at_port[:1])


async def test_handle_message_disconnected_many(
    subject: helpers.ModuleListener,
    mock_callback: AsyncMock,
    connections: List[models.ModuleConnection],
    modules_at_port: List[ModuleAtPort],
) -> None:
    """It should call the call back with the correct modules to remove."""
    message = models.Message(status="disconnected", connections=connections)
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with([], modules_at_port)


async def test_handle_message_dump_empty(
    subject: helpers.ModuleListener, mock_callback: AsyncMock
) -> None:
    """It should call the call back with the correct modules to load."""
    message = models.Message(status="dump", connections=[])
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with([], [])


async def test_handle_message_dump_one(
    subject: helpers.ModuleListener,
    mock_callback: AsyncMock,
    connections: List[models.ModuleConnection],
    modules_at_port: List[ModuleAtPort],
) -> None:
    """It should call the call back with the correct modules to load."""
    message = models.Message(status="dump", connections=connections[:1])
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with(modules_at_port[:1], [])


async def test_handle_message_dump_many(
    subject: helpers.ModuleListener,
    mock_callback: AsyncMock,
    connections: List[models.ModuleConnection],
    modules_at_port: List[ModuleAtPort],
) -> None:
    """It should call the call back with the correct modules to load."""
    message = models.Message(status="dump", connections=connections)
    await subject.handle_message(message=message)
    mock_callback.assert_called_once_with(modules_at_port, [])



"""Test temperature module core."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import TempDeck
from opentrons.hardware_control.modules.types import ModuleType, TemperatureStatus
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api.core.engine.module_core import TemperatureModuleCore
from opentrons.protocol_api.core.engine.protocol import ProtocolCore
from opentrons.protocol_api.core.engine.tasks import EngineTaskCore
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient

TempDeckHardware = SynchronousAdapter[TempDeck]


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> TempDeckHardware:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="TempDeckHardware")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def subject(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: TempDeckHardware,
    mock_protocol_core: ProtocolCore,
) -> TemperatureModuleCore:
    """Get a mock of TemperatureModuleCore."""
    return TemperatureModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: TempDeckHardware,
    mock_protocol_core: ProtocolCore,
) -> None:
    """It should be able to create a temperature module core."""
    result = TemperatureModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )

    assert result.module_id == "1234"
    assert result.MODULE_TYPE == ModuleType.TEMPERATURE


def test_set_target_temperature(
    decoy: Decoy,
    subject: TemperatureModuleCore,
    mock_engine_client: EngineClient,
) -> None:
    """Should verify EngineClient call to set_target_temperature and return an EngineTaskCore."""
    task_mock = decoy.mock(cls=EngineTaskCore)
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.temperature_module.SetTargetTemperatureParams(
                moduleId="1234", celsius=38.9
            )
        ),
    ).then_return(
        cmd.temperature_module.SetTargetTemperatureResult(
            targetTemperature=38.9, taskId="taskId"
        )
    )
    task_mock._id = "taskId"
    result = subject.set_target_temperature(38.9)
    assert isinstance(result, EngineTaskCore)
    assert result._id == "taskId"


def test_wait_for_target_temperature(
    decoy: Decoy,
    subject: TemperatureModuleCore,
    mock_engine_client: EngineClient,
) -> None:
    """Should verify EngineClient call to wait_for_target_temperature."""
    subject.wait_for_target_temperature()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.temperature_module.WaitForTemperatureParams(
                moduleId="1234", celsius=None
            )
        ),
        times=1,
    )


def test_deactivate(
    decoy: Decoy,
    subject: TemperatureModuleCore,
    mock_engine_client: EngineClient,
) -> None:
    """Should verify EngineClient call to deactivate temp module."""
    subject.deactivate()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.temperature_module.DeactivateTemperatureParams(moduleId="1234")
        ),
        times=1,
    )


def test_get_target_temperature(
    decoy: Decoy,
    subject: TemperatureModuleCore,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: TempDeckHardware,
) -> None:
    """Should return the target temperature."""
    decoy.when(mock_sync_module_hardware.target).then_return(38.9)

    result = subject.get_target_temperature()

    assert result == 38.9


def test_get_status(
    decoy: Decoy,
    subject: TemperatureModuleCore,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: TempDeckHardware,
) -> None:
    """Should get temp deck status."""
    decoy.when(mock_sync_module_hardware.status).then_return(TemperatureStatus.HEATING)

    assert subject.get_status() == TemperatureStatus.HEATING


def test_get_current_temperature(
    decoy: Decoy,
    subject: TemperatureModuleCore,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: TempDeckHardware,
) -> None:
    """Should get the current module temperature."""
    decoy.when(mock_sync_module_hardware.temperature).then_return(36.5)

    assert subject.get_current_temperature() == 36.5


def test_get_serial_number(
    decoy: Decoy, subject: TemperatureModuleCore, mock_engine_client: EngineClient
) -> None:
    """It should return a serial number."""
    decoy.when(mock_engine_client.state.modules.get_serial_number("1234")).then_return(
        "abc"
    )

    assert subject.get_serial_number() == "abc"



"""Utility functions and classes for the hardware controller"""

import asyncio
import logging
from enum import Enum
from typing import Any, Dict, List, Mapping, Optional, Tuple, TypeVar, Union

from opentrons_shared_data.errors.exceptions import MissingConfigurationData
from opentrons_shared_data.pipette.pipette_definition import (
    CamActionPickUpTipConfiguration,
    PipetteConfigurations,
    PressAndCamConfigurationValues,
    PressFitPickUpTipConfiguration,
)
from opentrons_shared_data.pipette.types import PipetteTipType

from .errors import OutOfBoundsMove
from .types import Axis, CriticalPoint, MotionChecks
from opentrons.types import Point

mod_log = logging.getLogger(__name__)


def _handle_loop_exception(
    loop: asyncio.AbstractEventLoop, context: Dict[str, Any]
) -> None:
    mod_log.error(f"Caught exception: {context}")


def use_or_initialize_loop(
    loop: Optional[asyncio.AbstractEventLoop],
) -> asyncio.AbstractEventLoop:
    checked_loop = loop or asyncio.get_event_loop()
    checked_loop.set_exception_handler(_handle_loop_exception)
    return checked_loop


def plan_arc(
    origin_point: Point,
    dest_point: Point,
    z_height: float,
    origin_cp: Optional[CriticalPoint] = None,
    dest_cp: Optional[CriticalPoint] = None,
    extra_waypoints: Optional[List[Tuple[float, float]]] = None,
) -> List[Tuple[Point, Optional[CriticalPoint]]]:
    assert z_height >= max(origin_point.z, dest_point.z)
    checked_wp = extra_waypoints or []
    return (
        [(origin_point._replace(z=z_height), origin_cp)]
        + [(Point(x=wp[0], y=wp[1], z=z_height), dest_cp) for wp in checked_wp]
        + [(dest_point._replace(z=z_height), dest_cp), (dest_point, dest_cp)]
    )


class DeckTransformState(Enum):
    OK = "OK"
    IDENTITY = "IDENTITY"
    BAD_CALIBRATION = "BAD_CALIBRATION"
    SINGULARITY = "SINGULARITY"

    def __str__(self) -> str:
        return self.name


AxisType = TypeVar("AxisType")


def check_motion_bounds(
    target_smoothie: Mapping[AxisType, float],
    target_deck: Mapping[AxisType, float],
    bounds: Mapping[AxisType, Tuple[float, float]],
    checks: MotionChecks,
) -> None:
    """
    Log or raise an error (depending on checks) if a specified
    target position is outside the acceptable bounds of motion

    Which axes are checked is defined by the elements of
    target_smoothie.

    The target_deck and mapping does not need to be
    entirely filled; it is used only for logging
    """
    bounds_message_format = (
        "Out of bounds move: {axis}=({tsp} motor controller, {tdp} deck) too "
        "{dir} for limit {limsp}"
    )
    for ax in target_smoothie.keys():
        if target_smoothie[ax] < bounds[ax][0]:
            format_detail = {
                "axis": ax,
                "tsp": target_smoothie[ax],
                "tdp": target_deck.get(ax, "unknown"),
                "dir": "low",
                "limsp": bounds.get(ax, ("unknown",))[0],
            }
            bounds_message = bounds_message_format.format(**format_detail)
            mod_log.warning(bounds_message)
            if checks.value & MotionChecks.LOW.value:
                raise OutOfBoundsMove(bounds_message, format_detail)
        elif target_smoothie[ax] > bounds[ax][1]:
            format_detail = {
                "axis": ax,
                "tsp": target_smoothie[ax],
                "tdp": target_deck.get(ax, "unknown"),
                "dir": "high",
                "limsp": bounds.get(ax, (None, "unknown"))[1],
            }
            bounds_message = bounds_message_format.format(**format_detail)
            mod_log.warning(bounds_message)
            if checks.value & MotionChecks.HIGH.value:
                raise OutOfBoundsMove(bounds_message, format_detail)


def ot2_axis_to_string(axis: Axis) -> str:
    """Returns OT-2 specific string for the given axis."""
    axis_str_map = {
        Axis.Z_L: "Z",
        Axis.Z_R: "A",
        Axis.P_L: "B",
        Axis.P_R: "C",
    }
    try:
        return axis_str_map[axis]
    except KeyError:
        return axis.name


def _get_press_and_cam_configuration_values(
    config: Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration],
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> PressAndCamConfigurationValues:
    try:
        return config.configuration_by_nozzle_map[valid_nozzle_map_key][
            active_tip_type.name
        ]
    except KeyError:
        default = config.configuration_by_nozzle_map[valid_nozzle_map_key].get(
            "default"
        )
        if default is not None:
            return default
        raise MissingConfigurationData(
            message=f"Default tip type configuration values do not exist for Nozzle Map {valid_nozzle_map_key}."
        )


def pick_up_speed_by_configuration(
    config: Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration],
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> float:
    """
    Returns the pick up Speed for a given configuration of a given pipette.

    Parameters:
        config: A PressFitPickUpTipConfiguration or a CamActionPickUpTipConfiguration (96ch Only) containing values for speed.
        valid_nozzle_map_key: The string key representing the current pre-validated map of nozzles the pipette has been configured to use.
        active_tip_type: The PipetteTipType to be attached to the pipette.
    """
    configuration_values = _get_press_and_cam_configuration_values(
        config, valid_nozzle_map_key, active_tip_type
    )
    return configuration_values.speed


def pick_up_distance_by_configuration(
    config: Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration],
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> float:
    """
    Returns the pick up Distance for a given configuration of a given pipette.

    Parameters:
        config: A PressFitPickUpTipConfiguration or a CamActionPickUpTipConfiguration (96ch Only) containing values for distance.
        valid_nozzle_map_key: The string key representing the current pre-validated map of nozzles the pipette has been configured to use.
        active_tip_type: The PipetteTipType to be attached to the pipette.
    """
    configuration_values = _get_press_and_cam_configuration_values(
        config, valid_nozzle_map_key, active_tip_type
    )
    return configuration_values.distance


def pick_up_current_by_configuration(
    config: Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration],
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> float:
    """
    Returns the pick up Current for a given configuration of a given pipette.

    Parameters:
        config: A PressFitPickUpTipConfiguration or a CamActionPickUpTipConfiguration (96ch Only) containing values for current.
        valid_nozzle_map_key: The string key representing the current pre-validated map of nozzles the pipette has been configured to use.
        active_tip_type: The PipetteTipType to be attached to the pipette.
    """
    configuration_values = _get_press_and_cam_configuration_values(
        config, valid_nozzle_map_key, active_tip_type
    )
    return configuration_values.current


def nominal_tip_overlap_dictionary_by_configuration(
    configurations: PipetteConfigurations,
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> Dict[str, Dict[str, float]]:
    """
    Returns the pick up speed for a given configuration of a given pipette.

    Parameters:
        configurations: The PipetterConfigurations of the pipette the tips will be attached to.
        valid_nozzle_map_key: The string key representing the current pre-validated map of nozzles the pipette has been configured to use.
        active_tip_type: The PipetteTipType to be attached to the pipette.
    """
    for config in (
        configurations.pick_up_tip_configurations.press_fit,
        configurations.pick_up_tip_configurations.cam_action,
    ):
        if not config:
            continue
        try:
            configuration_values = _get_press_and_cam_configuration_values(
                config, valid_nozzle_map_key, active_tip_type
            )
            return configuration_values.versioned_tip_overlap_dictionary
        except KeyError:
            # No valid key found for the approved nozzle map under this configuration - try the next
            continue
    raise MissingConfigurationData(
        message=f"No valid tip overlap dictionaries identified for map {valid_nozzle_map_key} using tip type {active_tip_type}."
    )



"""Build a module emulator for integration tests."""

import asyncio
from typing import Optional, Type, TypeVar, cast

from opentrons.drivers.rpi_drivers.types import PortGroup, USBPort
from opentrons.hardware_control import ExecutionManager
from opentrons.hardware_control.modules import AbstractModule

ModuleType = TypeVar("ModuleType", bound=AbstractModule)


async def build_module(
    module_cls: Type[ModuleType],
    *,
    port: int,
    execution_manager: ExecutionManager,
    poll_interval_seconds: Optional[float] = None,
) -> ModuleType:
    """Build a module emulator connected to a RFC-2217 network serial port.

    Args:
        module_cls: The module to build.
        port: The TCP port to connect on.
        execution_manager: Passed through to module builder.
        poll_interval_seconds: Passed through to module builder.

    Returns:
        The module's hardware control interface.
    """
    module = await module_cls.build(
        port=f"socket://127.0.0.1:{port}",
        execution_manager=execution_manager,
        disconnected_callback=lambda *args: None,
        error_callback=lambda *args: None,
        usb_port=USBPort(
            name="",
            port_number=1,
            port_group=PortGroup.UNKNOWN,
            device_path="",
            hub=False,
            hub_port=None,
        ),
        hw_control_loop=asyncio.get_running_loop(),
        poll_interval_seconds=poll_interval_seconds,
    )

    return cast(ModuleType, module)



"""Test labware movement command execution side effects."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.labware.labware_definition import LabwareDefinition2

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import Axis, OT3Mount
from opentrons.protocol_engine.errors import (
    GripperNotAttachedError,
    HardwareNotSupportedError,
    HeaterShakerLabwareLatchNotOpenError,
    LabwareMovementNotAllowedError,
    ThermocyclerNotOpenError,
)
from opentrons.protocol_engine.execution import EquipmentHandler, MovementHandler
from opentrons.protocol_engine.execution.heater_shaker_movement_flagger import (
    HeaterShakerMovementFlagger,
)
from opentrons.protocol_engine.execution.labware_movement import (
    LabwareMovementHandler,
)
from opentrons.protocol_engine.execution.thermocycler_movement_flagger import (
    ThermocyclerMovementFlagger,
)
from opentrons.protocol_engine.execution.thermocycler_plate_lifter import (
    ThermocyclerPlateLifter,
)
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    Dimensions,
    GripperMoveType,
    GripSpecs,
    LabwareLocation,
    LabwareOffset,
    LabwareOffsetVector,
    LegacyLabwareOffsetLocation,
    ModuleLocation,
    NonStackedLocation,
    OnLabwareLocation,
)
from opentrons.types import DeckSlotName, Point

if TYPE_CHECKING:
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mocked out StateStore instance."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def equipment(decoy: Decoy) -> EquipmentHandler:
    """Get a mocked out EquipmentHandler instance."""
    return decoy.mock(cls=EquipmentHandler)


@pytest.fixture
def movement(decoy: Decoy) -> MovementHandler:
    """Get a mocked out MovementHandler."""
    return decoy.mock(cls=MovementHandler)


@pytest.fixture
def thermocycler_plate_lifter(decoy: Decoy) -> ThermocyclerPlateLifter:
    """Get a mocked out ThermocyclerPlateLifter instance."""
    return decoy.mock(cls=ThermocyclerPlateLifter)


@pytest.fixture
def thermocycler_movement_flagger(decoy: Decoy) -> ThermocyclerMovementFlagger:
    """Get a mocked out ThermocyclerMovementFlagger instance."""
    return decoy.mock(cls=ThermocyclerMovementFlagger)


@pytest.fixture
def heater_shaker_movement_flagger(decoy: Decoy) -> HeaterShakerMovementFlagger:
    """Get a mocked out HeaterShakerMovementFlagger instance."""
    return decoy.mock(cls=HeaterShakerMovementFlagger)


@pytest.fixture
def labware_def(decoy: Decoy) -> LabwareDefinition2:
    """Get a mocked out LabwareDefinition2 instance."""
    return decoy.mock(cls=LabwareDefinition2)


async def set_up_decoy_hardware_gripper(
    decoy: Decoy,
    ot3_hardware_api: OT3API,
    state_store: StateStore,
    labware_def: LabwareDefinition2,
) -> None:
    """Shared hardware gripper decoy setup."""
    decoy.when(state_store.config.use_virtual_gripper).then_return(False)
    decoy.when(ot3_hardware_api.has_gripper()).then_return(True)
    decoy.when(ot3_hardware_api.gripper_jaw_can_home()).then_return(True)
    assert ot3_hardware_api.hardware_gripper
    decoy.when(
        await ot3_hardware_api.gantry_position(mount=OT3Mount.GRIPPER)
    ).then_return(Point(x=777, y=888, z=999))

    decoy.when(
        await ot3_hardware_api.encoder_current_position_ot3(OT3Mount.GRIPPER)
    ).then_return({Axis.G: 4.0})

    decoy.when(
        ot3_hardware_api.hardware_gripper.geometry.max_allowed_grip_error
    ).then_return(6.0)

    decoy.when(ot3_hardware_api.hardware_gripper.jaw_width).then_return(89)

    decoy.when(state_store.labware.get_grip_force(labware_def)).then_return(100)

    decoy.when(state_store.labware.get_labware_offset("new-offset-id")).then_return(
        LabwareOffset(
            id="new-offset-id",
            createdAt=datetime(year=2022, month=10, day=20),
            definitionUri="my-labware",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_5
            ),  # this location doesn't matter for this test
            vector=LabwareOffsetVector(x=0.5, y=0.6, z=0.7),
        )
    )


@pytest.fixture
def subject(
    ot3_hardware_api: OT3API,
    state_store: StateStore,
    equipment: EquipmentHandler,
    movement: MovementHandler,
    thermocycler_plate_lifter: ThermocyclerPlateLifter,
    thermocycler_movement_flagger: ThermocyclerMovementFlagger,
    heater_shaker_movement_flagger: HeaterShakerMovementFlagger,
) -> LabwareMovementHandler:
    """Get LabwareMovementHandler for OT3, with its dependencies mocked out."""
    return LabwareMovementHandler(
        hardware_api=ot3_hardware_api,
        state_store=state_store,
        equipment=equipment,
        movement=movement,
        thermocycler_plate_lifter=thermocycler_plate_lifter,
        thermocycler_movement_flagger=thermocycler_movement_flagger,
        heater_shaker_movement_flagger=heater_shaker_movement_flagger,
    )


@pytest.mark.ot3_only
async def test_raise_error_if_gripper_pickup_failed(
    decoy: Decoy,
    state_store: StateStore,
    thermocycler_plate_lifter: ThermocyclerPlateLifter,
    ot3_hardware_api: OT3API,
    labware_def: LabwareDefinition2,
    subject: LabwareMovementHandler,
) -> None:
    """Test that the gripper position check is called at the right time."""
    #  This function should only be called when after the gripper opens,
    #  and then closes again. This is when we expect the labware to be
    #  in the gripper jaws.
    await set_up_decoy_hardware_gripper(
        decoy, ot3_hardware_api, state_store, labware_def
    )
    assert ot3_hardware_api.hardware_gripper

    user_pick_up_offset = Point(x=123, y=234, z=345)
    user_drop_offset = Point(x=111, y=222, z=333)

    starting_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    to_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_2)

    labware_def.parameters.quirks = []

    decoy.when(
        state_store.labware.get_definition("my-teleporting-labware")
    ).then_return(labware_def)

    mock_tc_context_manager = decoy.mock(name="mock_tc_context_manager")
    decoy.when(
        thermocycler_plate_lifter.lift_plate_for_labware_movement(
            labware_location=starting_location
        )
    ).then_return(mock_tc_context_manager)

    decoy.when(
        state_store.geometry.get_labware_grip_point(
            labware_definition=labware_def,
            location=starting_location,
            move_type=GripperMoveType.PICK_UP_LABWARE,
            user_additional_offset=user_pick_up_offset,
        )
    ).then_return(Point(101, 102, 119.5))

    decoy.when(
        state_store.geometry.get_labware_grip_point(
            labware_definition=labware_def,
            location=to_location,
            move_type=GripperMoveType.DROP_LABWARE,
            user_additional_offset=user_drop_offset,
        )
    ).then_return(Point(201, 202, 219.5))

    decoy.when(
        state_store.labware.get_gripper_width_specs(labware_definition=labware_def)
    ).then_return(GripSpecs(targetY=100, uncertaintyNarrower=5, uncertaintyWider=10))

    await subject.move_labware_with_gripper(
        labware_id="my-teleporting-labware",
        current_location=starting_location,
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_2),
        user_pick_up_offset=user_pick_up_offset,
        user_drop_offset=user_drop_offset,
        post_drop_slide_offset=Point(x=1, y=1, z=1),
    )

    decoy.verify(
        await ot3_hardware_api.home(axes=[Axis.Z_L, Axis.Z_R, Axis.Z_G]),
        await mock_tc_context_manager.__aenter__(),
        await ot3_hardware_api.grip(force_newtons=100),
        await ot3_hardware_api.ungrip(),
        await ot3_hardware_api.grip(force_newtons=100),
        ot3_hardware_api.raise_error_if_gripper_pickup_failed(
            expected_grip_width=100,
            grip_width_uncertainty_wider=10,
            grip_width_uncertainty_narrower=5,
            disable_geometry_grip_check=False,
        ),
        await ot3_hardware_api.grip(force_newtons=100),
        ot3_hardware_api.raise_error_if_gripper_pickup_failed(
            expected_grip_width=100,
            grip_width_uncertainty_wider=10,
            grip_width_uncertainty_narrower=5,
            disable_geometry_grip_check=False,
        ),
        await ot3_hardware_api.grip(force_newtons=100),
        ot3_hardware_api.raise_error_if_gripper_pickup_failed(
            expected_grip_width=100,
            grip_width_uncertainty_wider=10,
            grip_width_uncertainty_narrower=5,
            disable_geometry_grip_check=False,
        ),
        await ot3_hardware_api.disengage_axes([Axis.Z_G]),
        await ot3_hardware_api.ungrip(),
        await ot3_hardware_api.ungrip(),
    )


# TODO (spp, 2022-10-18):
#  1. Should write an acceptance test w/ real labware on ot3 deck.
#  2. This test will be split once waypoints generation is moved to motion planning.
@pytest.mark.parametrize(
    argnames=["from_location", "to_location", "slide_offset"],
    argvalues=[
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
            None,
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            ModuleLocation(moduleId="module-id"),
            Point(x=50, y=0, z=0),
        ),
        (
            OnLabwareLocation(labwareId="a-labware-id"),
            OnLabwareLocation(labwareId="another-labware-id"),
            None,
        ),
        (
            ModuleLocation(moduleId="a-module-id"),
            DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            Point(x=-10, y=40, z=-40),
        ),
    ],
)
@pytest.mark.ot3_only
async def test_move_labware_with_gripper(
    decoy: Decoy,
    state_store: StateStore,
    thermocycler_plate_lifter: ThermocyclerPlateLifter,
    ot3_hardware_api: OT3API,
    subject: LabwareMovementHandler,
    labware_def: LabwareDefinition2,
    from_location: Union[DeckSlotLocation, ModuleLocation, OnLabwareLocation],
    to_location: Union[DeckSlotLocation, ModuleLocation, OnLabwareLocation],
    slide_offset: Optional[Point],
) -> None:
    """It should perform a labware movement with gripper by delegating to OT3API."""
    # TODO (spp, 2023-07-26): this test does NOT stub out movement waypoints in order to
    #  keep this as the semi-smoke test that it previously was. We should add a proper
    #  smoke test for gripper labware movement with actual labware and make this a unit test.
    labware_def.parameters.quirks = []
    await set_up_decoy_hardware_gripper(
        decoy, ot3_hardware_api, state_store, labware_def
    )

    decoy.when(
        state_store.labware.get_definition("my-teleporting-labware")
    ).then_return(labware_def)

    user_pick_up_offset = Point(x=123, y=234, z=345)
    user_drop_offset = Point(x=111, y=222, z=333)

    decoy.when(
        state_store.labware.get_dimensions(labware_definition=labware_def)
    ).then_return(Dimensions(x=100, y=85, z=0))

    decoy.when(
        state_store.labware.get_well_bbox(labware_definition=labware_def)
    ).then_return(Dimensions(x=99, y=80, z=1))

    pickup_grip_point = Point(101, 102, 119.5)
    drop_grip_point = Point(201, 202, 219.5)

    decoy.when(
        state_store.geometry.get_labware_grip_point(
            labware_definition=labware_def,
            location=from_location,
            move_type=GripperMoveType.PICK_UP_LABWARE,
            user_additional_offset=user_pick_up_offset,
        )
    ).then_return(pickup_grip_point)

    decoy.when(
        state_store.geometry.get_labware_grip_point(
            labware_definition=labware_def,
            location=to_location,
            move_type=GripperMoveType.DROP_LABWARE,
            user_additional_offset=user_drop_offset,
        )
    ).then_return(drop_grip_point)

    mock_tc_context_manager = decoy.mock(name="mock_tc_context_manager")
    decoy.when(
        thermocycler_plate_lifter.lift_plate_for_labware_movement(
            labware_location=from_location
        )
    ).then_return(mock_tc_context_manager)

    decoy.when(
        state_store.labware.get_gripper_width_specs(labware_definition=labware_def)
    ).then_return(GripSpecs(targetY=100, uncertaintyNarrower=5, uncertaintyWider=10))

    expected_waypoints = [
        Point(
            pickup_grip_point.x, pickup_grip_point.y, 999
        ),  # move to above pickup location
        pickup_grip_point,  # move to pickup location
        Point(
            pickup_grip_point.x, pickup_grip_point.y, 999
        ),  # gripper retract at pickup location
        Point(drop_grip_point.x, drop_grip_point.y, 999),  # move to above drop location
        drop_grip_point,  # move down to drop location
        Point(drop_grip_point.x, drop_grip_point.y, 999),  # retract at drop location
    ]

    await subject.move_labware_with_gripper(
        labware_id="my-teleporting-labware",
        current_location=from_location,
        new_location=to_location,
        user_pick_up_offset=user_pick_up_offset,
        user_drop_offset=user_drop_offset,
        post_drop_slide_offset=slide_offset,
    )

    gripper = OT3Mount.GRIPPER

    if slide_offset is None:
        decoy.verify(
            await ot3_hardware_api.home(axes=[Axis.Z_L, Axis.Z_R, Axis.Z_G]),
            await mock_tc_context_manager.__aenter__(),
            await ot3_hardware_api.grip(force_newtons=100),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[0]
            ),
            await ot3_hardware_api.ungrip(),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[1]
            ),
            await ot3_hardware_api.grip(force_newtons=100),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[2]
            ),
            await ot3_hardware_api.grip(force_newtons=100),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[3]
            ),
            await ot3_hardware_api.grip(force_newtons=100),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[4]
            ),
            await ot3_hardware_api.disengage_axes([Axis.Z_G]),
            await ot3_hardware_api.ungrip(),
            await ot3_hardware_api.home_z(OT3Mount.GRIPPER),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[5]
            ),
            await ot3_hardware_api.idle_gripper(),
        )
    else:
        decoy.verify(
            await ot3_hardware_api.home(axes=[Axis.Z_L, Axis.Z_R, Axis.Z_G]),
            await mock_tc_context_manager.__aenter__(),
            await ot3_hardware_api.grip(force_newtons=100),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[0]
            ),
            await ot3_hardware_api.ungrip(),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[1]
            ),
            await ot3_hardware_api.grip(force_newtons=100),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[2]
            ),
            await ot3_hardware_api.grip(force_newtons=100),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[3]
            ),
            await ot3_hardware_api.grip(force_newtons=100),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[4]
            ),
            await ot3_hardware_api.disengage_axes([Axis.Z_G]),
            await ot3_hardware_api.ungrip(),
            await ot3_hardware_api.home_z(OT3Mount.GRIPPER),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[5]
            ),
            await ot3_hardware_api.ungrip(),
            await ot3_hardware_api.move_to(
                mount=gripper, abs_position=expected_waypoints[5] + slide_offset
            ),
            await ot3_hardware_api.idle_gripper(),
        )


async def test_labware_movement_raises_on_ot2(
    decoy: Decoy,
    state_store: StateStore,
    hardware_api: HardwareControlAPI,
    equipment: EquipmentHandler,
    movement: MovementHandler,
) -> None:
    """It should raise an error when attempting a gripper movement on a non-OT3 bot."""
    decoy.when(state_store.config.use_virtual_gripper).then_return(False)
    subject = LabwareMovementHandler(
        hardware_api=hardware_api,
        state_store=state_store,
        equipment=equipment,
        movement=movement,
    )

    with pytest.raises(HardwareNotSupportedError):
        await subject.move_labware_with_gripper(
            labware_id="labware-id",
            current_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
            new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            user_pick_up_offset=Point(),
            user_drop_offset=Point(),
            post_drop_slide_offset=None,
        )


@pytest.mark.ot3_only
async def test_labware_movement_skips_for_virtual_gripper(
    decoy: Decoy,
    state_store: StateStore,
    ot3_hardware_api: OT3API,
    subject: LabwareMovementHandler,
) -> None:
    """It should neither raise error nor move gripper when using virtual gripper."""
    decoy.when(state_store.config.use_virtual_gripper).then_return(True)
    await subject.move_labware_with_gripper(
        labware_id="labware-id",
        current_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        user_pick_up_offset=Point(),
        user_drop_offset=Point(),
        post_drop_slide_offset=None,
    )
    decoy.verify(
        await ot3_hardware_api.move_to(
            mount=matchers.Anything(), abs_position=matchers.Anything()
        ),
        times=0,
        ignore_extra_args=True,
    )


@pytest.mark.ot3_only
async def test_labware_movement_raises_without_gripper(
    decoy: Decoy,
    state_store: StateStore,
    ot3_hardware_api: OT3API,
    subject: LabwareMovementHandler,
) -> None:
    """It should raise an error when attempting a gripper movement without a gripper."""
    decoy.when(state_store.config.use_virtual_gripper).then_return(False)
    decoy.when(ot3_hardware_api.has_gripper()).then_return(False)

    with pytest.raises(GripperNotAttachedError):
        await subject.move_labware_with_gripper(
            labware_id="labware-id",
            current_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
            new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            user_pick_up_offset=Point(),
            user_drop_offset=Point(),
            post_drop_slide_offset=None,
        )


@pytest.mark.parametrize(
    argnames=["from_loc", "to_loc"],
    argvalues=[
        (
            ModuleLocation(moduleId="a-thermocycler-id"),
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
            ModuleLocation(moduleId="a-thermocycler-id"),
        ),
    ],
)
async def test_ensure_movement_obstructed_by_thermocycler_raises(
    decoy: Decoy,
    subject: LabwareMovementHandler,
    state_store: StateStore,
    thermocycler_movement_flagger: ThermocyclerMovementFlagger,
    from_loc: NonStackedLocation,
    to_loc: LabwareLocation,
) -> None:
    """It should raise error when labware movement is obstructed by thermocycler."""
    decoy.when(
        state_store.labware.get_parent_location(labware_id="labware-id")
    ).then_return(from_loc)
    decoy.when(
        await thermocycler_movement_flagger.ensure_labware_in_open_thermocycler(  # type: ignore[func-returns-value]
            labware_parent=ModuleLocation(moduleId="a-thermocycler-id")
        )
    ).then_raise(ThermocyclerNotOpenError("Thou shall not pass!"))
    decoy.when(state_store.labware.is_lid(labware_id="labware-id")).then_return(False)
    with pytest.raises(LabwareMovementNotAllowedError):
        await subject.ensure_movement_not_obstructed_by_module(
            labware_id="labware-id", new_location=to_loc
        )


async def test_ensure_movement_not_obstructed_by_modules(
    decoy: Decoy,
    subject: LabwareMovementHandler,
    state_store: StateStore,
) -> None:
    """It should not raise error when labware movement is not obstructed by thermocycler."""
    decoy.when(
        state_store.labware.get_parent_location(labware_id="labware-id")
    ).then_return(ModuleLocation(moduleId="a-rando-module-id"))
    decoy.when(state_store.labware.is_lid(labware_id="labware-id")).then_return(False)
    await subject.ensure_movement_not_obstructed_by_module(
        labware_id="labware-id",
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
    )


@pytest.mark.parametrize(
    argnames=["from_loc", "to_loc"],
    argvalues=[
        (
            ModuleLocation(moduleId="a-heater-shaker-id"),
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
            ModuleLocation(moduleId="a-heater-shaker-id"),
        ),
    ],
)
async def test_ensure_movement_obstructed_by_heater_shaker_raises(
    decoy: Decoy,
    subject: LabwareMovementHandler,
    heater_shaker_movement_flagger: HeaterShakerMovementFlagger,
    state_store: StateStore,
    from_loc: NonStackedLocation,
    to_loc: LabwareLocation,
) -> None:
    """It should raise error when labware movement is obstructed by thermocycler."""
    decoy.when(
        state_store.labware.get_parent_location(labware_id="labware-id")
    ).then_return(from_loc)
    decoy.when(state_store.labware.is_lid(labware_id="labware-id")).then_return(False)
    decoy.when(
        await heater_shaker_movement_flagger.raise_if_labware_latched_on_heater_shaker(
            labware_parent=ModuleLocation(moduleId="a-heater-shaker-id")
        )
    ).then_raise(HeaterShakerLabwareLatchNotOpenError("Thou shall not take!"))

    with pytest.raises(LabwareMovementNotAllowedError):
        await subject.ensure_movement_not_obstructed_by_module(
            labware_id="labware-id", new_location=to_loc
        )


async def test_ensure_movement_not_obstructed_does_not_raise_for_slot_locations(
    decoy: Decoy,
    subject: LabwareMovementHandler,
    state_store: StateStore,
) -> None:
    """It should not raise error when moving from slot to slot."""
    decoy.when(
        state_store.labware.get_parent_location(labware_id="labware-id")
    ).then_return(DeckSlotLocation(slotName=DeckSlotName.SLOT_1))
    await subject.ensure_movement_not_obstructed_by_module(
        labware_id="labware-id",
        new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
    )


@pytest.mark.parametrize(
    argnames=["from_loc", "to_loc"],
    argvalues=[
        (
            ModuleLocation(moduleId="a-heater-shaker-id"),
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
            ModuleLocation(moduleId="a-heater-shaker-id"),
        ),
    ],
)
async def test_ensure_movement_not_obstructed_does_not_raise_for_lids_on_hs(
    decoy: Decoy,
    subject: LabwareMovementHandler,
    heater_shaker_movement_flagger: HeaterShakerMovementFlagger,
    state_store: StateStore,
    from_loc: NonStackedLocation,
    to_loc: LabwareLocation,
) -> None:
    """It should not raise any errors when moving lid from a latched H/S."""
    decoy.when(
        state_store.labware.get_parent_location(labware_id="labware-id")
    ).then_return(from_loc)
    decoy.when(state_store.labware.is_lid(labware_id="labware-id")).then_return(True)
    decoy.when(
        await heater_shaker_movement_flagger.raise_if_labware_latched_on_heater_shaker(
            labware_parent=ModuleLocation(moduleId="a-heater-shaker-id")
        )
    ).then_raise(HeaterShakerLabwareLatchNotOpenError("Thou shall not take!"))
    await subject.ensure_movement_not_obstructed_by_module(
        labware_id="labware-id", new_location=to_loc
    )



"""Test comment command."""

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.comment import (
    CommentImplementation,
    CommentParams,
    CommentResult,
)


async def test_comment_implementation() -> None:
    """Confirm that comment command can be created and executed."""
    subject = CommentImplementation()

    data = CommentParams(message="hello world")

    result = await subject.execute(data)

    assert result == SuccessData(public=CommentResult())



"""Tests for the protocol engine's ActionDispatcher."""

from decoy import Decoy

from opentrons.protocol_engine.actions import (
    ActionDispatcher,
    ActionHandler,
    StopAction,
)


def test_sink(decoy: Decoy) -> None:
    """It should send all actions to the sink handler."""
    action = StopAction()

    sink = decoy.mock(cls=ActionHandler)
    subject = ActionDispatcher(sink=sink)

    subject.dispatch(action)

    decoy.verify(sink.handle_action(action))


def test_add_handler(decoy: Decoy) -> None:
    """It should actions to handlers before the sink."""
    action = StopAction()

    handler_1 = decoy.mock(cls=ActionHandler)
    handler_2 = decoy.mock(cls=ActionHandler)
    sink = decoy.mock(cls=ActionHandler)

    subject = ActionDispatcher(sink=sink)
    subject.add_handler(handler_1)
    subject.add_handler(handler_2)
    subject.dispatch(action)

    decoy.verify(
        handler_1.handle_action(action),
        handler_2.handle_action(action),
        sink.handle_action(action),
    )



import pytest

from opentrons.protocol_api.module_validation_and_errors import (
    InvalidTargetSpeedError,
    InvalidTargetTemperatureError,
    validate_heater_shaker_speed,
    validate_heater_shaker_temperature,
)
from opentrons.protocols.api_support.types import APIVersion


@pytest.mark.parametrize("valid_celsius_value", [37.0, 37.1, 50, 94.99, 95])
def test_validate_heater_shaker_temperature(valid_celsius_value: float) -> None:
    """It should return the validated temperature value."""
    validated = validate_heater_shaker_temperature(
        celsius=valid_celsius_value, api_version=APIVersion(2, 25)
    )
    assert validated == valid_celsius_value


@pytest.mark.parametrize("invalid_celsius_value", [-1, 0, 36.99, 95.01])
def test_validate_heater_shaker_temperature_raises_under_api_225(
    invalid_celsius_value: float,
) -> None:
    """It should raise an error for invalid temperature values."""
    with pytest.raises(InvalidTargetTemperatureError):
        validate_heater_shaker_temperature(
            celsius=invalid_celsius_value, api_version=APIVersion(2, 24)
        )


@pytest.mark.parametrize("invalid_celsius_value", [-1, 95.01])
def test_validate_heater_shaker_temperature_raises_over_api_225(
    invalid_celsius_value: float,
) -> None:
    """It should raise an error for invalid temperature values."""
    with pytest.raises(InvalidTargetTemperatureError):
        validate_heater_shaker_temperature(
            celsius=invalid_celsius_value, api_version=APIVersion(2, 25)
        )


def test_validate_heater_shaker_temperature_passes_low_temp_over_api_225() -> None:
    """It should not raise an error for values under 37."""
    assert validate_heater_shaker_temperature(celsius=22, api_version=APIVersion(2, 25))


@pytest.mark.parametrize("valid_rpm_value", [200, 201, 1000, 2999, 3000])
def test_validate_heater_shaker_speed(valid_rpm_value: int) -> None:
    """It should return the validated speed value."""
    validated = validate_heater_shaker_speed(rpm=valid_rpm_value)
    assert validated == valid_rpm_value


@pytest.mark.parametrize("invalid_rpm_value", [0, 199, 3001])
def test_validate_heater_shaker_speed_raises(invalid_rpm_value: int) -> None:
    """It should raise an error for invalid speed values."""
    with pytest.raises(InvalidTargetSpeedError):
        validate_heater_shaker_speed(rpm=invalid_rpm_value)



"""Command models to manually retrieve a labware from a Flex Stacker in an unsafe situation."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Union, cast

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Type

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from ...errors import (
    CannotPerformModuleAction,
    ErrorOccurrence,
    FlexStackerLabwarePoolNotYetDefinedError,
    LocationIsOccupiedError,
)
from ...resources import ModelUtils
from ...state import update_types
from ...types import (
    InStackerHopperLocation,
    LabwareLocationSequence,
    ModuleLocation,
)
from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..flex_stacker.common import (
    FlexStackerStallOrCollisionError,
    adapter_location_sequence,
    build_retrieve_labware_move_updates,
    labware_locations_for_group,
    lid_location_sequence,
    primary_location_sequence,
)
from opentrons.hardware_control.modules.types import PlatformState

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler
    from opentrons.protocol_engine.state.state import StateView

UnsafeFlexStackerManualRetrieveCommandType = Literal[
    "unsafe/flexStacker/manualRetrieve"
]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class UnsafeFlexStackerManualRetrieveParams(BaseModel):
    """Input parameters for a labware retrieval command."""

    moduleId: str = Field(
        ...,
        description="Unique ID of the Flex Stacker.",
    )
    labwareId: str | SkipJsonSchema[None] = Field(
        None,
        description="Do not use. Present for internal backward compatibility.",
        json_schema_extra=_remove_default,
    )
    displayName: str | SkipJsonSchema[None] = Field(
        None,
        description="Do not use. Present for internal backward compatibility.",
        json_schema_extra=_remove_default,
    )
    adapterId: str | SkipJsonSchema[None] = Field(
        None,
        description="Do not use. Present for internal backward compatibility.",
        json_schema_extra=_remove_default,
    )
    lidId: str | SkipJsonSchema[None] = Field(
        None,
        description="Do not use. Present for internal backward compatibility.",
        json_schema_extra=_remove_default,
    )


class UnsafeFlexStackerManualRetrieveResult(BaseModel):
    """Result data from a labware retrieval command."""

    labwareId: str = Field(
        ...,
        description="The labware ID of the primary retrieved labware.",
    )
    adapterId: str | None = Field(
        None,
        description="The optional Adapter Labware ID of the adapter under a primary labware.",
    )
    lidId: str | None = Field(
        None,
        description="The optional Lid Labware ID of the lid on a primary labware.",
    )
    primaryLocationSequence: LabwareLocationSequence = Field(
        ..., description="The origin location of the primary labware."
    )
    lidLocationSequence: LabwareLocationSequence | None = Field(
        None,
        description="The origin location of the adapter labware under a primary labware.",
    )
    adapterLocationSequence: LabwareLocationSequence | None = Field(
        None, description="The origin location of the lid labware on a primary labware."
    )
    originalPrimaryLocationSequence: LabwareLocationSequence = Field(
        ..., description="The original location of the just-retrieved primary labware"
    )
    originalAdapterLocationSequence: LabwareLocationSequence | None = Field(
        None, description="The original location of an adapter labware if any"
    )
    originalLidLocationSequence: LabwareLocationSequence | None = Field(
        None, description="The original location of a lid labware if any"
    )
    primaryLabwareURI: str = Field(
        ...,
        description="The labware definition URI of the primary labware.",
    )
    adapterLabwareURI: str | None = Field(
        None,
        description="The labware definition URI of the adapter labware.",
    )
    lidLabwareURI: str | None = Field(
        None,
        description="The labware definition URI of the lid labware.",
    )


_ExecuteReturn = Union[
    SuccessData[UnsafeFlexStackerManualRetrieveResult],
    DefinedErrorData[FlexStackerStallOrCollisionError],
]


class UnsafeFlexStackerManualRetrieveImpl(
    AbstractCommandImpl[UnsafeFlexStackerManualRetrieveParams, _ExecuteReturn]
):
    """Implementation of a labware retrieval command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._model_utils = model_utils

    async def execute(
        self, params: UnsafeFlexStackerManualRetrieveParams
    ) -> _ExecuteReturn:
        """Execute the labware retrieval command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )

        pool_definitions = stacker_state.get_pool_definition_ordered_list()
        location = self._state_view.modules.get_location(params.moduleId)
        if pool_definitions is None:
            raise FlexStackerLabwarePoolNotYetDefinedError(
                message=f"The Flex Stacker in {location} has not been configured yet and cannot be filled."
            )

        if len(stacker_state.contained_labware_bottom_first) == 0:
            raise CannotPerformModuleAction(
                message=f"Cannot retrieve labware from Flex Stacker in {location} because it contains no labware"
            )

        stacker_loc = ModuleLocation(moduleId=params.moduleId)
        # Allow propagation of ModuleNotAttachedError.
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        # Validate that the stacker is fully in the gripper position
        if stacker_hw:
            stacker_hw.set_stacker_identify(True)
            if stacker_hw.platform_state != PlatformState.EXTENDED:
                raise CannotPerformModuleAction(
                    f"Cannot manually retrieve a labware from Flex Stacker in {location} if the carriage is not in gripper position."
                )

        try:
            # In theory given this is an unsafe manual retrieve this should never raise
            self._state_view.labware.raise_if_labware_in_location(stacker_loc)
        except LocationIsOccupiedError:
            raise CannotPerformModuleAction(
                f"Cannot retrieve a labware from Flex Stacker in {location} if the carriage is occupied"
            )

        to_retrieve = stacker_state.contained_labware_bottom_first[0]
        remaining = stacker_state.contained_labware_bottom_first[1:]

        locations, offsets = build_retrieve_labware_move_updates(
            to_retrieve, stacker_state, self._state_view
        )

        # Update the state to reflect the labware is now in the Flex Stacker slot
        # todo(chb, 2025-02-19): This ModuleLocation piece should probably instead be an AddressableAreaLocation
        # but that has implications for where labware are set by things like module.load_labware(..) and what
        # happens when we move labware.
        stacker_area = (
            self._state_view.modules.ensure_and_convert_module_fixture_location(
                deck_slot=self._state_view.modules.get_location(
                    params.moduleId
                ).slotName,
                model=self._state_view.modules.get(params.moduleId).model,
            )
        )

        original_locations = labware_locations_for_group(
            to_retrieve,
            self._state_view.geometry.get_predicted_location_sequence(
                InStackerHopperLocation(moduleId=params.moduleId)
            ),
        )
        new_locations = labware_locations_for_group(
            to_retrieve,
            self._state_view.geometry.get_predicted_location_sequence(
                ModuleLocation(moduleId=params.moduleId)
            ),
        )

        if stacker_hw is not None:
            stacker_hw.set_stacker_identify(False)

        return SuccessData(
            public=UnsafeFlexStackerManualRetrieveResult(
                labwareId=to_retrieve.primaryLabwareId,
                adapterId=to_retrieve.adapterLabwareId,
                lidId=to_retrieve.lidLabwareId,
                primaryLocationSequence=primary_location_sequence(new_locations),
                adapterLocationSequence=adapter_location_sequence(new_locations),
                lidLocationSequence=lid_location_sequence(new_locations),
                originalPrimaryLocationSequence=primary_location_sequence(
                    original_locations
                ),
                originalAdapterLocationSequence=adapter_location_sequence(
                    original_locations
                ),
                originalLidLocationSequence=lid_location_sequence(original_locations),
                primaryLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    cast(LabwareDefinition, stacker_state.pool_primary_definition)
                ),
                adapterLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_adapter_definition
                ),
                lidLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_lid_definition
                ),
            ),
            state_update=(
                update_types.StateUpdate()
                .set_batch_labware_location(
                    new_locations_by_id=locations, new_offset_ids_by_id=offsets
                )
                .update_flex_stacker_contained_labware(params.moduleId, remaining)
                .set_addressable_area_used(stacker_area)
            ),
        )


class UnsafeFlexStackerManualRetrieve(
    BaseCommand[
        UnsafeFlexStackerManualRetrieveParams,
        UnsafeFlexStackerManualRetrieveResult,
        ErrorOccurrence,
    ]
):
    """A command to manually retrieve a labware from a Flex Stacker."""

    commandType: UnsafeFlexStackerManualRetrieveCommandType = (
        "unsafe/flexStacker/manualRetrieve"
    )
    params: UnsafeFlexStackerManualRetrieveParams
    result: UnsafeFlexStackerManualRetrieveResult | None = None

    _ImplementationCls: Type[UnsafeFlexStackerManualRetrieveImpl] = (
        UnsafeFlexStackerManualRetrieveImpl
    )


class UnsafeFlexStackerManualRetrieveCreate(
    BaseCommandCreate[UnsafeFlexStackerManualRetrieveParams]
):
    """A request to execute a Flex Stacker manual retrieve command."""

    commandType: UnsafeFlexStackerManualRetrieveCommandType = (
        "unsafe/flexStacker/manualRetrieve"
    )
    params: UnsafeFlexStackerManualRetrieveParams

    _CommandCls: Type[UnsafeFlexStackerManualRetrieve] = UnsafeFlexStackerManualRetrieve



"""Tests for the APIs around liquid classes."""

from typing import Any, Dict

import pytest

from opentrons_shared_data.liquid_classes.liquid_class_definition import Coordinate
from opentrons_shared_data.liquid_classes.types import TransferPropertiesDict

from opentrons.protocol_api import ProtocolContext


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_liquid_class_creation_and_property_fetching(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should create the liquid class and provide access to its properties."""
    pipette_load_name = "flex_8channel_50"
    p50 = simulated_protocol_context.load_instrument(pipette_load_name, mount="left")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    assert water.name == "water"
    assert water.display_name == "Aqueous"

    # TODO (spp, 2024-10-17): update this to fetch pipette load name from instrument context
    assert (
        water.get_for(p50, tiprack).dispense.flow_rate_by_volume.get_for_volume(1) == 50
    )
    assert water.get_for(pipette_load_name, tiprack.uri).aspirate.submerge.speed == 100

    with pytest.raises(
        ValueError,
        match="No properties found for non-existent-pipette in water liquid class",
    ):
        water.get_for("non-existent-pipette", tiprack.uri)

    with pytest.raises(AttributeError):
        water.name = "foo"  # type: ignore

    with pytest.raises(AttributeError):
        water.display_name = "bar"  # type: ignore

    with pytest.raises(ValueError, match="Liquid class definition not found"):
        simulated_protocol_context.get_liquid_class("non-existent-liquid")


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_custom_liquid_class_creation_and_property_fetching(
    simulated_protocol_context: ProtocolContext,
    minimal_transfer_properties_dict: Dict[str, Dict[str, TransferPropertiesDict]],
) -> None:
    """It should create the liquid class and provide access to its properties."""
    pipette_load_name = "flex_1channel_50"
    p50 = simulated_protocol_context.load_instrument(pipette_load_name, mount="left")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    custom_water = simulated_protocol_context.define_liquid_class(
        name="water_50",
        properties=minimal_transfer_properties_dict,
        display_name="Custom Aqueous",
    )
    custom_water_props = custom_water.get_for(p50, tiprack)
    assert custom_water_props.aspirate.submerge.speed == 100
    assert custom_water_props.dispense.flow_rate_by_volume.get_for_volume(20) == 30

    with pytest.raises(
        ValueError,
        match="No properties found for flex_8channel_50 in water_50 liquid class",
    ):
        custom_water.get_for(
            "flex_8channel_50", "opentrons/opentrons_flex_96_tiprack_50ul/1"
        )


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_custom_liquid_class_w_multi_dispense_creation_and_property_fetching(
    simulated_protocol_context: ProtocolContext,
    custom_pip_n_tip_transfer_properties_dict_v2: Dict[str, Dict[str, Any]],
) -> None:
    """It should create the liquid class with specified properties and provide access to its properties."""
    p50 = "a_custom_pipette_type"
    tiprack = "a_custom_tiprack_uri"
    custom_water = simulated_protocol_context.define_liquid_class(
        name="water_50",
        properties=custom_pip_n_tip_transfer_properties_dict_v2,
        display_name="Custom Aqueous",
    )

    assert custom_water.name == "water_50"
    assert custom_water.display_name == "Custom Aqueous"
    custom_water_props = custom_water.get_for(p50, tiprack)
    assert custom_water_props.aspirate.retract.speed == 40
    assert custom_water_props.dispense.submerge.speed == 100
    assert custom_water_props.multi_dispense.retract.blowout.location.value == "trash"  # type: ignore[union-attr]

    with pytest.raises(
        ValueError,
        match="No properties found for flex_8channel_50 in water_50 liquid class",
    ):
        custom_water.get_for(
            "flex_8channel_50", "opentrons/opentrons_flex_96_tiprack_50ul/1"
        )


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.28", "Flex")], indirect=True
)
def test_adding_blowout_position_to_liquid_classes(
    simulated_protocol_context: ProtocolContext,
    custom_pip_n_tip_transfer_properties_dict_v2: Dict[str, Dict[str, Any]],
) -> None:
    """It should accept a blowout position in liquid class properties."""
    pip_type = "a_custom_pipette_type"
    tiprack = "a_custom_tiprack_uri"
    custom_pip_n_tip_transfer_properties_dict_v2[pip_type][tiprack]["dispense"][
        "retract"
    ]["blowout"] = {
        "enable": True,
        "location": "source",
        "flow_rate": 321,
        "blowout_position": {
            "position_reference": "well-top",
            "offset": {"x": 11, "y": 22, "z": 33},
        },
    }
    custom_water = simulated_protocol_context.define_liquid_class(
        name="water_50",
        properties=custom_pip_n_tip_transfer_properties_dict_v2,
        display_name="Custom Aqueous",
    )
    custom_water_props = custom_water.get_for(pip_type, tiprack)
    assert custom_water_props.dispense.retract.blowout.blowout_position is not None
    assert (
        custom_water_props.dispense.retract.blowout.blowout_position.position_reference.value
        == "well-top"
    )
    assert (
        custom_water_props.dispense.retract.blowout.blowout_position.offset
        == Coordinate(x=11, y=22, z=33)
    )

    custom_water_props.dispense.retract.blowout.blowout_position = None
    assert custom_water_props.dispense.retract.blowout.blowout_position is None

    with pytest.raises(ValueError):
        custom_water_props.dispense.retract.blowout.blowout_position = {  # type:ignore[assignment]
            "position_reference": "well-top",
            "offset": {"x": 11, "y": 22, "z": 33, "w": 123},
        }



"""Tests for Protocol API thermocycler module contexts."""

import inspect
from typing import Optional

import pytest
from decoy import Decoy, matchers

from opentrons.drivers.types import ThermocyclerLidStatus
from opentrons.hardware_control.modules import TemperatureStatus
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import (
    MAX_SUPPORTED_VERSION,
    ThermocyclerContext,
)
from opentrons.protocol_api import (
    validation as mock_validation,
)
from opentrons.protocol_api.core.common import (
    LabwareCore,
    ProtocolCore,
    ThermocyclerCore,
)
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocol_api.labware import Labware, Well
from opentrons.protocol_api.tasks import Task
from opentrons.protocols.api_support.types import APIVersion


@pytest.fixture(autouse=True)
def _mock_validation_module(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(mock_validation, inspect.isfunction):
        monkeypatch.setattr(mock_validation, name, decoy.mock(func=func))


@pytest.fixture
def mock_core(decoy: Decoy) -> ThermocyclerCore:
    """Get a mock module implementation core."""
    return decoy.mock(cls=ThermocyclerCore)


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol implementation core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock LoadedCoreMap."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def mock_broker(decoy: Decoy) -> LegacyBroker:
    """Get a mock command message broker."""
    return decoy.mock(cls=LegacyBroker)


@pytest.fixture
def mock_labware(decoy: Decoy) -> Labware:
    """Get a mock command message broker."""
    return decoy.mock(cls=Labware)


@pytest.fixture
def mock_well(decoy: Decoy) -> Well:
    """Get a mock command message broker."""
    return decoy.mock(cls=Well)


@pytest.fixture
def api_version() -> APIVersion:
    """Get an API version to apply to the interface."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def subject(
    api_version: APIVersion,
    mock_core: ThermocyclerCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_broker: LegacyBroker,
) -> ThermocyclerContext:
    """Get a thermocycler module context with its dependencies mocked out."""
    return ThermocyclerContext(
        core=mock_core,
        protocol_core=mock_protocol_core,
        core_map=mock_core_map,
        broker=mock_broker,
        api_version=api_version,
    )


def test_get_lid_position(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the lid position status from the core."""
    decoy.when(mock_core.get_lid_position()).then_return(ThermocyclerLidStatus.OPEN)
    result = subject.lid_position

    assert result == "open"
    assert isinstance(result, ThermocyclerLidStatus) is False


def test_get_block_temperature_status(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the block temperature status from the core."""
    decoy.when(mock_core.get_block_temperature_status()).then_return(
        TemperatureStatus.IDLE
    )
    result = subject.block_temperature_status

    assert result == "idle"
    assert isinstance(result, TemperatureStatus) is False


def test_get_lid_temperature_status(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the lid temperature status from the core."""
    decoy.when(mock_core.get_lid_temperature_status()).then_return(
        TemperatureStatus.IDLE
    )
    result = subject.lid_temperature_status

    assert result == "idle"
    assert isinstance(result, TemperatureStatus) is False


def test_get_block_temperature(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the current block temperature from the core."""
    decoy.when(mock_core.get_block_temperature()).then_return(12.3)
    result = subject.block_temperature
    assert result == 12.3


def test_get_block_target_temperature(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the target block temperature from the core."""
    decoy.when(mock_core.get_block_target_temperature()).then_return(12.3)
    result = subject.block_target_temperature
    assert result == 12.3


def test_get_lid_temperature(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the current lid temperature from the core."""
    decoy.when(mock_core.get_lid_temperature()).then_return(42.0)
    result = subject.lid_temperature
    assert result == 42.0


def test_get_lid_target_temperature(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the target lid temperature from the core."""
    decoy.when(mock_core.get_lid_target_temperature()).then_return(42.0)
    result = subject.lid_target_temperature
    assert result == 42.0


def test_get_ramp_rate(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the current ramp rate from the core."""
    decoy.when(mock_core.get_ramp_rate()).then_return(1.23)
    result = subject.ramp_rate
    assert result == 1.23


def test_get_hold_time(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the current hold time from the core."""
    decoy.when(mock_core.get_hold_time()).then_return(13.37)
    result = subject.hold_time
    assert result == 13.37


def test_get_total_cycle_count(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the total cycle count from the core."""
    decoy.when(mock_core.get_total_cycle_count()).then_return(321)
    result = subject.total_cycle_count
    assert result == 321


def test_get_current_cycle_index(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the current cycle index from the core."""
    decoy.when(mock_core.get_current_cycle_index()).then_return(123)
    result = subject.current_cycle_index
    assert result == 123


def test_get_total_step_count(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the total step count from the core."""
    decoy.when(mock_core.get_total_step_count()).then_return(1337)
    result = subject.total_step_count
    assert result == 1337


def test_get_current_step_index(
    decoy: Decoy, mock_core: ThermocyclerCore, subject: ThermocyclerContext
) -> None:
    """It should get the current step index from the core."""
    decoy.when(mock_core.get_current_step_index()).then_return(42)
    result = subject.current_step_index
    assert result == 42


def test_open_lid(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should open the lid via the core."""
    decoy.when(mock_core.open_lid()).then_return(ThermocyclerLidStatus.OPEN)

    result = subject.open_lid()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_OPEN",
                }
            ),
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )

    assert result == "open"
    assert isinstance(result, ThermocyclerLidStatus) is False


def test_close_lid(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should close the lid via the core."""
    decoy.when(mock_core.close_lid()).then_return(ThermocyclerLidStatus.CLOSED)

    result = subject.close_lid()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_CLOSE",
                }
            ),
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )

    assert result == "closed"
    assert isinstance(result, ThermocyclerLidStatus) is False


def test_start_set_block_temperature(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
    api_version: APIVersion,
) -> None:
    """It should set the block temperature via the core."""
    subject._api_version = api_version
    mock_task = decoy.mock(cls=Task)
    decoy.when(
        mock_core.start_set_target_block_temperature(
            celsius=42.0,
            ramp_rate=5.6,
            block_max_volume=7.8,
        )
    ).then_return(mock_task._core)
    result = subject.start_set_block_temperature(
        temperature=42.0,
        block_max_volume=7.8,
        ramp_rate=5.6,
    )

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_START_SET_BLOCK_TEMP",
                    "payload": matchers.DictMatching({"temperature": 42.0}),
                }
            ),
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "after",
                    "payload": matchers.DictMatching({"temperature": 42.0}),
                }
            ),
        ),
    )
    assert isinstance(result, Task)
    assert result._core is mock_task._core
    assert result._api_version == api_version


def test_set_block_temperature(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should set the block temperature via the core."""
    decoy.when(
        mock_validation.ensure_hold_time_seconds(seconds=1.2, minutes=3.4)
    ).then_return(205.2)

    subject.set_block_temperature(
        temperature=42.0,
        hold_time_seconds=1.2,
        hold_time_minutes=3.4,
        ramp_rate=5.6,
        block_max_volume=7.8,
    )

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_SET_BLOCK_TEMP",
                    "payload": matchers.DictMatching(
                        {"temperature": 42.0, "hold_time": 205.2}
                    ),
                }
            ),
        ),
        mock_core.set_target_block_temperature(
            celsius=42.0,
            hold_time_seconds=205.2,
            block_max_volume=7.8,
            ramp_rate=5.6,
        ),
        mock_core.wait_for_block_temperature(),
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "after",
                    "payload": matchers.DictMatching(
                        {"temperature": 42.0, "hold_time": 205.2}
                    ),
                }
            ),
        ),
    )


def test_get_current_labware_max_volume(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_labware: Labware,
    mock_well: Well,
    subject: ThermocyclerContext,
) -> None:
    """It should return a max block volume within bounds."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_well.has_tracked_liquid()).then_return(True)
    decoy.when(mock_well.current_liquid_volume()).then_return(125.0)
    decoy.when(mock_labware.wells()).then_return([mock_well])
    decoy.when(subject._protocol_core.get_labware_on_module(mock_core)).then_return(
        mock_labware_core
    )
    decoy.when(subject._core_map.get(mock_labware_core)).then_return(mock_labware)
    result = subject._get_current_labware_max_vol()
    assert result == 100.0
    decoy.when(mock_well.current_liquid_volume()).then_return(-10.0)
    result = subject._get_current_labware_max_vol()
    assert result == 0.0


def test_set_block_temperature_with_liquid_tracking(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    mock_labware: Labware,
    mock_well: Well,
    subject: ThermocyclerContext,
) -> None:
    """It should set the block temperature via the core."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_well.has_tracked_liquid()).then_return(True)
    decoy.when(mock_well.current_liquid_volume()).then_return(42.0)
    decoy.when(mock_labware.wells()).then_return([mock_well])
    decoy.when(subject._protocol_core.get_labware_on_module(mock_core)).then_return(
        mock_labware_core
    )
    decoy.when(subject._core_map.get(mock_labware_core)).then_return(mock_labware)

    decoy.when(
        mock_validation.ensure_hold_time_seconds(seconds=1.2, minutes=3.4)
    ).then_return(5.6)

    subject.set_block_temperature(
        temperature=42.0,
        hold_time_seconds=1.2,
        hold_time_minutes=3.4,
        ramp_rate=5.6,
    )

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_SET_BLOCK_TEMP",
                    "payload": matchers.DictMatching(
                        {"temperature": 42.0, "hold_time": 205.2}
                    ),
                }
            ),
        ),
        mock_core.set_target_block_temperature(
            celsius=42.0,
            hold_time_seconds=5.6,
            block_max_volume=42,
            ramp_rate=5.6,
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "after",
                    "payload": matchers.DictMatching(
                        {"temperature": 42.0, "hold_time": 205.2}
                    ),
                }
            ),
        ),
    )


def test_set_lid_temperature(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should close the lid via the core."""
    subject.set_lid_temperature(42.0)
    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_SET_LID_TEMP",
                }
            ),
        ),
        mock_core.set_target_lid_temperature(celsius=42.0),
        mock_core.wait_for_lid_temperature(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_start_set_lid_temperature(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should close the lid via the core."""
    mock_task = decoy.mock(cls=Task)
    decoy.when(mock_core.start_set_target_lid_temperature(42.0)).then_return(
        mock_task._core
    )
    result = subject.start_set_lid_temperature(42.0)
    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_START_SET_LID_TEMP",
                }
            ),
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )
    assert isinstance(result, Task)
    assert result._core is mock_task._core


def test_start_execute_profile(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
    api_version: APIVersion,
) -> None:
    """It should execute a thermocycler profile via the core and return a task."""
    subject._api_version = api_version
    mock_task = decoy.mock(cls=Task)
    decoy.when(mock_validation.ensure_thermocycler_repetition_count(123)).then_return(
        321
    )
    decoy.when(
        mock_validation.ensure_thermocycler_profile_steps(
            [
                {
                    "temperature": 12.3,
                    "hold_time_minutes": 12.3,
                    "hold_time_seconds": 45.6,
                }
            ]
        )
    ).then_return([{"temperature": 42.0, "hold_time_seconds": 123.456}])
    decoy.when(
        mock_core.start_execute_profile(
            steps=[
                {
                    "temperature": 42.0,
                    "hold_time_seconds": 123.456,
                }
            ],
            repetitions=321,
            block_max_volume=34.5,
        )
    ).then_return(mock_task._core)
    result = subject.start_execute_profile(
        steps=[
            {"temperature": 12.3, "hold_time_minutes": 12.3, "hold_time_seconds": 45.6}
        ],
        repetitions=123,
        block_max_volume=34.5,
    )

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_START_EXECUTE_PROFILE",
                    "payload": matchers.DictMatching(
                        {
                            "steps": [
                                {
                                    "temperature": 12.3,
                                    "hold_time_minutes": 12.3,
                                    "hold_time_seconds": 45.6,
                                }
                            ]
                        }
                    ),
                }
            ),
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "after",
                    "payload": matchers.DictMatching(
                        {
                            "steps": [
                                {
                                    "temperature": 12.3,
                                    "hold_time_minutes": 12.3,
                                    "hold_time_seconds": 45.6,
                                }
                            ]
                        }
                    ),
                }
            ),
        ),
    )
    assert isinstance(result, Task)
    assert result._core is mock_task._core
    assert result._api_version == api_version


def test_execute_profile(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should execute a thermocycler profile via the core."""
    decoy.when(mock_validation.ensure_thermocycler_repetition_count(123)).then_return(
        321
    )
    decoy.when(
        mock_validation.ensure_thermocycler_profile_steps(
            [
                {
                    "temperature": 12.3,
                    "hold_time_minutes": 12.3,
                    "hold_time_seconds": 45.6,
                }
            ]
        )
    ).then_return([{"temperature": 42.0, "hold_time_seconds": 123.456}])

    subject.execute_profile(
        steps=[
            {"temperature": 12.3, "hold_time_minutes": 12.3, "hold_time_seconds": 45.6}
        ],
        repetitions=123,
        block_max_volume=34.5,
    )

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_EXECUTE_PROFILE",
                    "payload": matchers.DictMatching(
                        {
                            "steps": [
                                {
                                    "temperature": 12.3,
                                    "hold_time_minutes": 12.3,
                                    "hold_time_seconds": 45.6,
                                }
                            ]
                        }
                    ),
                }
            ),
        ),
        mock_core.execute_profile(
            steps=[
                {
                    "temperature": 42.0,
                    "hold_time_seconds": 123.456,
                }
            ],
            repetitions=321,
            block_max_volume=34.5,
        ),
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "after",
                    "payload": matchers.DictMatching(
                        {
                            "steps": [
                                {
                                    "temperature": 12.3,
                                    "hold_time_minutes": 12.3,
                                    "hold_time_seconds": 45.6,
                                }
                            ]
                        }
                    ),
                }
            ),
        ),
    )


def test_deactivate_lid(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should turn off the heated lid via the core."""
    subject.deactivate_lid()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_DEACTIVATE_LID",
                }
            ),
        ),
        mock_core.deactivate_lid(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_deactivate_block(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should turn off the well block temperature controller via the core."""
    subject.deactivate_block()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_DEACTIVATE_BLOCK",
                }
            ),
        ),
        mock_core.deactivate_block(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


def test_deactivate(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    mock_broker: LegacyBroker,
    subject: ThermocyclerContext,
) -> None:
    """It should turn off the well block and heated lid via the core."""
    subject.deactivate()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.THERMOCYCLER_DEACTIVATE",
                }
            ),
        ),
        mock_core.deactivate(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


@pytest.mark.parametrize(
    "this_api_version,ramp_rate", [(APIVersion(2, 27), None), (APIVersion(2, 28), 3.0)]
)
def test_ramp_rate_(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    mock_core: ThermocyclerCore,
    subject: ThermocyclerContext,
    this_api_version: APIVersion,
    ramp_rate: Optional[float],
) -> None:
    """It should delete the ramp rate argument on pre 2.28 versions."""
    subject._api_version = this_api_version
    decoy.when(
        mock_validation.ensure_hold_time_seconds(seconds=None, minutes=None)
    ).then_return(0)
    decoy.when(subject._get_current_labware_max_vol()).then_return(None)
    subject.set_block_temperature(temperature=100, ramp_rate=3.0)

    decoy.verify(
        mock_core.set_target_block_temperature(
            celsius=100,
            hold_time_seconds=0,
            block_max_volume=None,
            ramp_rate=ramp_rate,
        )
    )


def test_serial_number(
    decoy: Decoy,
    mock_core: ThermocyclerCore,
    subject: ThermocyclerContext,
) -> None:
    """It should get the module's unique serial number."""
    decoy.when(mock_core.get_serial_number()).then_return("abc-123")
    result = subject.serial_number
    assert result == "abc-123"



"""Protocol engine types for labware offset vectors.

This is a separate module to avoid circular imports.
"""

from __future__ import annotations

from pydantic import BaseModel


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class LabwareOffsetVector(BaseModel):
    """Offset, in deck coordinates from nominal to actual position."""

    x: float
    y: float
    z: float



"""Test the loadLid command."""

import inspect

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.load_lid import (
    LoadLidImplementation,
    LoadLidParams,
    LoadLidResult,
)
from opentrons.protocol_engine.execution import EquipmentHandler, LoadedLabwareData
from opentrons.protocol_engine.resources import labware_validation
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    LabwareLidUpdate,
    LoadedLabwareUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    OnAddressableAreaLocationSequenceComponent,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
)


@pytest.fixture(autouse=True)
def patch_mock_labware_validation(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out labware_validations.py functions."""
    for name, func in inspect.getmembers(labware_validation, inspect.isfunction):
        monkeypatch.setattr(labware_validation, name, decoy.mock(func=func))


async def test_load_lid_on_tiprack(
    decoy: Decoy,
    tiprack_lid_def: LabwareDefinition,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """It should load a lid onto a labware."""
    subject = LoadLidImplementation(equipment=equipment, state_view=state_view)
    input_location = OnLabwareLocation(labwareId="some-labware-id")
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(location=input_location)
    ).then_return(input_location)
    decoy.when(
        await equipment.load_labware(
            load_name="someLid",
            namespace="someNamespace",
            version=1,
            location=input_location,
            labware_id=None,
        )
    ).then_return(
        LoadedLabwareData(
            labware_id="labware-id",
            definition=tiprack_lid_def,
            offsetId=None,
        )
    )

    decoy.when(
        labware_validation.validate_definition_is_lid(tiprack_lid_def)
    ).then_return(True)

    decoy.when(
        state_view.labware.raise_if_labware_cannot_be_stacked(
            top_labware_definition=tiprack_lid_def,
            bottom_labware_id=input_location.labwareId,
        )
    ).then_return(True)
    decoy.when(
        state_view.geometry.get_location_sequence(input_location.labwareId)
    ).then_return(
        [OnAddressableAreaLocationSequenceComponent(addressableAreaName="someAa")]
    )

    result = await subject.execute(
        LoadLidParams(
            location=input_location,
            loadName="someLid",
            namespace="someNamespace",
            version=1,
        )
    )
    assert result == SuccessData(
        public=LoadLidResult(
            definition=tiprack_lid_def,
            labwareId="labware-id",
            locationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId=input_location.labwareId, lidId="labware-id"
                ),
                OnAddressableAreaLocationSequenceComponent(
                    addressableAreaName="someAa"
                ),
            ],
        ),
        state_update=StateUpdate(
            labware_lid=LabwareLidUpdate(
                parent_labware_ids=["some-labware-id"], lid_ids=["labware-id"]
            ),
            loaded_labware=LoadedLabwareUpdate(
                definition=tiprack_lid_def,
                labware_id="labware-id",
                offset_id=None,
                display_name=None,
                new_location=input_location,
            ),
        ),
    )



"""Partial tip configurations for tests."""

from contextlib import nullcontext as does_not_raise
from typing import Any, ContextManager, NamedTuple, Optional

import pytest

from opentrons.protocol_api._nozzle_layout import NozzleLayout
from opentrons.protocols.api_support.util import UnsupportedAPIError


class NozzleLayoutArgs(NamedTuple):
    """Arguments to use in `configure_nozzle_layout`."""

    style: NozzleLayout
    start: Optional[str]
    end: Optional[str]
    front_right: Optional[str]
    back_left: Optional[str]


class PipetteIndependentNozzleConfigSpec(NamedTuple):
    """Parametrization data for pipette-independent nozzle configs test."""

    nozzle_layout_args: NozzleLayoutArgs
    expected_raise: ContextManager[Any]


PIPETTE_INDEPENDENT_TEST_SPECS = [
    PipetteIndependentNozzleConfigSpec(
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.SINGLE,
            start=None,
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Cannot configure a nozzle layout of style SINGLE without a starting nozzle.",
        ),
    ),
    PipetteIndependentNozzleConfigSpec(
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.SINGLE,
            start="H1",
            end="B1",
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with the SINGLE nozzle configuration.",
        ),
    ),
    PipetteIndependentNozzleConfigSpec(
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.SINGLE,
            start="H1",
            end=None,
            front_right="C1",
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with the SINGLE nozzle configuration.",
        ),
    ),
    PipetteIndependentNozzleConfigSpec(
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.SINGLE,
            start="H1",
            end=None,
            front_right=None,
            back_left="C1",
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with the SINGLE nozzle configuration.",
        ),
    ),
    PipetteIndependentNozzleConfigSpec(
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.SINGLE,
            start="A1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=does_not_raise(),
    ),
    PipetteIndependentNozzleConfigSpec(
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ALL,
            start="H1",
            end="G1",
            front_right="X1",
            back_left="Z1",
        ),
        expected_raise=does_not_raise(),  # Arguments are ignored
    ),
    PipetteIndependentNozzleConfigSpec(
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ALL,
            start=None,
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=does_not_raise(),
    ),
    PipetteIndependentNozzleConfigSpec(
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.QUADRANT,
            start="H1",
            end="G1",
            front_right="X1",
            back_left="Z1",
        ),
        expected_raise=pytest.raises(UnsupportedAPIError),
    ),
]


class PipetteReliantNozzleConfigSpec(NamedTuple):
    """Test parametrization data."""

    pipette_channels: int
    nozzle_layout_args: NozzleLayoutArgs
    expected_raise: ContextManager[Any]


PIPETTE_RELIANT_TEST_SPECS = [
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.COLUMN,
            start="A1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="COLUMN configuration is only supported on 96-Channel pipettes.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.COLUMN,
            start="E1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Starting nozzle specified is not one of \\['A1', 'H1', 'A12', 'H12'\\].",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.COLUMN,
            start="A1",
            end="B1",
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with"
            " the COLUMN nozzle configuration.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.COLUMN,
            start="A1",
            end=None,
            front_right="A1",
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with"
            " the COLUMN nozzle configuration.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.COLUMN,
            start="A1",
            end=None,
            front_right=None,
            back_left="B1",
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with"
            " the COLUMN nozzle configuration.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.COLUMN,
            start="A1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=does_not_raise(),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ROW,
            start="H1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="ROW configuration is only supported on 96-Channel pipettes.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ROW,
            start="B1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Starting nozzle specified is not one of \\['A1', 'H1', 'A12', 'H12'\\].",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ROW,
            start="A1",
            end="B1",
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with"
            " the ROW nozzle configuration.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ROW,
            start="A1",
            end=None,
            front_right="A1",
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with"
            " the ROW nozzle configuration.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ROW,
            start="A1",
            end=None,
            front_right=None,
            back_left="B1",
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'end', 'front_right' and 'back_left' cannot be used with"
            " the ROW nozzle configuration.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ROW,
            start="H1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_raise=does_not_raise(),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="H1",
            end="G1",
            front_right=None,
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Partial column configuration is only supported on 8-Channel pipettes",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="H1",
            end=None,
            front_right="H1",
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Partial column configurations require the 'end' parameter.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="H1",
            end="G1",
            front_right="H1",
            back_left=None,
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'front_right' and 'back_left' cannot be used with the PARTIAL_COLUMN configuration.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="H1",
            end="G1",
            front_right=None,
            back_left="H1",
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="Parameters 'front_right' and 'back_left' cannot be used with the PARTIAL_COLUMN configuration.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="H1",
            end="A1",
            front_right=None,
            back_left="H1",
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="A partial column configuration with 'start'=H1 cannot have its 'end' parameter be in row A."
            " Use `ALL` configuration to utilize all nozzles.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="A1",
            end="H1",
            front_right=None,
            back_left="H1",
        ),
        expected_raise=pytest.raises(
            ValueError,
            match="A partial column configuration with 'start'=A1 cannot have its 'end' parameter be in row H."
            " Use `ALL` configuration to utilize all nozzles.",
        ),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="A1",
            end="G1",
            front_right=None,
            back_left=None,
        ),
        expected_raise=does_not_raise(),
    ),
    PipetteReliantNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="H1",
            end="G1",
            front_right=None,
            back_left=None,
        ),
        expected_raise=does_not_raise(),
    ),
]


class ExpectedCoreArgs(NamedTuple):
    """The converted arguments expected to be passed to instrument core."""

    primary_nozzle: Optional[str]
    front_right_nozzle: Optional[str]
    back_left_nozzle: Optional[str]


class InstrumentCoreNozzleConfigSpec(NamedTuple):
    """Test parametrization data."""

    pipette_channels: int
    nozzle_layout_args: NozzleLayoutArgs
    expected_core_args: ExpectedCoreArgs


INSTRUMENT_CORE_NOZZLE_LAYOUT_TEST_SPECS = [
    InstrumentCoreNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.SINGLE,
            start="A1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_core_args=ExpectedCoreArgs(
            primary_nozzle="A1",
            front_right_nozzle=None,
            back_left_nozzle=None,
        ),
    ),
    InstrumentCoreNozzleConfigSpec(
        pipette_channels=8,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.PARTIAL_COLUMN,
            start="A1",
            end="D1",
            front_right=None,
            back_left=None,
        ),
        expected_core_args=ExpectedCoreArgs(
            primary_nozzle="A1",
            front_right_nozzle="D1",
            back_left_nozzle="A1",
        ),
    ),
    InstrumentCoreNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.COLUMN,
            start="H1",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_core_args=ExpectedCoreArgs(
            primary_nozzle="H1",
            front_right_nozzle=None,
            back_left_nozzle=None,
        ),
    ),
    InstrumentCoreNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ROW,
            start="A12",
            end=None,
            front_right=None,
            back_left=None,
        ),
        expected_core_args=ExpectedCoreArgs(
            primary_nozzle="A12",
            front_right_nozzle=None,
            back_left_nozzle=None,
        ),
    ),
    InstrumentCoreNozzleConfigSpec(
        pipette_channels=96,
        nozzle_layout_args=NozzleLayoutArgs(
            style=NozzleLayout.ALL,
            start="A12",
            end=None,
            front_right="D3",
            back_left=None,
        ),
        expected_core_args=ExpectedCoreArgs(  # These args are eventually ignored
            primary_nozzle="A12",
            front_right_nozzle="D3",
            back_left_nozzle=None,
        ),
    ),
]



"""Test pipette seal commands."""

from datetime import datetime
from unittest.mock import sentinel

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import StallOrCollisionDetectedError
from opentrons_shared_data.labware import load_definition
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
    labware_definition_type_adapter,
)

from opentrons.hardware_control import OT3HardwareControlAPI
from opentrons.protocol_engine import (
    DeckPoint,
    PickUpTipWellLocation,
    WellLocation,
    WellOffset,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.commands.seal_pipette_to_tip import (
    SealPipetteToTipImplementation,
    SealPipetteToTipParams,
    SealPipetteToTipResult,
)
from opentrons.protocol_engine.errors import PickUpTipTipNotAttachedError
from opentrons.protocol_engine.execution import (
    GantryMover,
    MovementHandler,
    PipettingHandler,
    TipHandler,
)
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import (
    AspiratedFluid,
    FluidKind,
    LabwareWellId,
    TipGeometry,
)
from opentrons.types import MountType, Point


@pytest.fixture
def evotips_definition() -> LabwareDefinition:
    """A fixturee of the evotips definition."""
    # TODO (chb 2025-01-29): When we migrate all labware to v3 we can clean this up
    return labware_definition_type_adapter.validate_python(
        load_definition("ev_resin_tips_flex_96_labware", 1)
    )


async def test_success(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    tip_handler: TipHandler,
    model_utils: ModelUtils,
    gantry_mover: GantryMover,
    evotips_definition: LabwareDefinition,
    ot3_hardware_api: OT3HardwareControlAPI,
    pipetting: PipettingHandler,
) -> None:
    """A PickUpTip command should have an execution implementation."""
    subject = SealPipetteToTipImplementation(
        state_view=state_view,
        movement=movement,
        tip_handler=tip_handler,
        model_utils=model_utils,
        gantry_mover=gantry_mover,
        hardware_api=ot3_hardware_api,
        pipetting=pipetting,
    )

    decoy.when(state_view.pipettes.get_mount("pipette-id")).then_return(MountType.LEFT)

    decoy.when(
        state_view.geometry.convert_pick_up_tip_well_location(
            well_location=PickUpTipWellLocation(offset=WellOffset(x=1, y=2, z=3))
        )
    ).then_return(WellLocation(offset=WellOffset(x=1, y=2, z=3)))

    decoy.when(state_view.labware.get_definition("labware-id")).then_return(
        evotips_definition
    )
    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="A3",
            well_location=WellLocation(offset=WellOffset(x=1, y=2, z=3)),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=111, y=222, z=333))
    decoy.when(
        state_view.geometry.get_nominal_tip_geometry("pipette-id", "labware-id", "A3")
    ).then_return(TipGeometry(length=42, diameter=5, volume=300))

    decoy.when(
        await tip_handler.pick_up_tip(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="A3",
        )
    ).then_return(TipGeometry(length=42, diameter=5, volume=300))

    result = await subject.execute(
        SealPipetteToTipParams(
            pipetteId="pipette-id",
            labwareId="labware-id",
            wellName="A3",
            wellLocation=PickUpTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        )
    )

    assert result == SuccessData(
        public=SealPipetteToTipResult(
            tipLength=42,
            tipVolume=300,
            tipDiameter=5,
            position=DeckPoint(x=111, y=222, z=333),
        ),
        state_update=update_types.StateUpdate(
            pipette_tip_state=update_types.PipetteTipStateUpdate(
                pipette_id="pipette-id",
                tip_geometry=TipGeometry(length=42, diameter=5, volume=300),
                tip_source=LabwareWellId(labware_id="labware-id", well_name="A3"),
            ),
            pipette_aspirated_fluid=update_types.PipetteAspiratedFluidUpdate(
                pipette_id="pipette-id",
                fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=1000),
            ),
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(labware_id="labware-id", well_name="A3"),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
        ),
    )


async def test_no_tip_physically_missing_error(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    tip_handler: TipHandler,
    model_utils: ModelUtils,
    gantry_mover: GantryMover,
    ot3_hardware_api: OT3HardwareControlAPI,
    pipetting: PipettingHandler,
    evotips_definition: LabwareDefinition,
) -> None:
    """It should not return a TipPhysicallyMissingError even though evotips do not sit high enough on the pipette to be detected by the tip sensor."""
    subject = SealPipetteToTipImplementation(
        state_view=state_view,
        movement=movement,
        tip_handler=tip_handler,
        model_utils=model_utils,
        gantry_mover=gantry_mover,
        hardware_api=ot3_hardware_api,
        pipetting=pipetting,
    )

    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    error_id = "error-id"
    error_created_at = datetime(1234, 5, 6)

    decoy.when(
        state_view.geometry.convert_pick_up_tip_well_location(
            well_location=PickUpTipWellLocation(offset=WellOffset())
        )
    ).then_return(WellLocation(offset=WellOffset()))

    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(offset=WellOffset()),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=111, y=222, z=333))
    decoy.when(
        state_view.geometry.get_nominal_tip_geometry(pipette_id, labware_id, well_name)
    ).then_return(TipGeometry(length=42, diameter=5, volume=300))

    decoy.when(
        await tip_handler.pick_up_tip(
            pipette_id=pipette_id, labware_id=labware_id, well_name=well_name
        )
    ).then_raise(PickUpTipTipNotAttachedError(tip_geometry=sentinel.tip_geometry))
    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_created_at)
    decoy.when(state_view.labware.get_definition(labware_id)).then_return(
        evotips_definition
    )

    result = await subject.execute(
        SealPipetteToTipParams(
            pipetteId=pipette_id, labwareId=labware_id, wellName=well_name
        )
    )

    assert result == SuccessData(
        public=SealPipetteToTipResult(
            tipLength=42,
            tipVolume=300,
            tipDiameter=5,
            position=DeckPoint(x=111, y=222, z=333),
        ),
        state_update=update_types.StateUpdate(
            pipette_tip_state=update_types.PipetteTipStateUpdate(
                pipette_id="pipette-id",
                tip_geometry=TipGeometry(length=42, diameter=5, volume=300),
                tip_source=LabwareWellId(
                    labware_id="labware-id", well_name="well-name"
                ),
            ),
            pipette_aspirated_fluid=update_types.PipetteAspiratedFluidUpdate(
                pipette_id="pipette-id",
                fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=1000),
            ),
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(
                    labware_id="labware-id", well_name="well-name"
                ),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
        ),
    )


async def test_stall_error(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    tip_handler: TipHandler,
    model_utils: ModelUtils,
    gantry_mover: GantryMover,
    ot3_hardware_api: OT3HardwareControlAPI,
    pipetting: PipettingHandler,
    evotips_definition: LabwareDefinition,
) -> None:
    """It should return a TipPhysicallyMissingError if the HW API indicates that."""
    subject = SealPipetteToTipImplementation(
        state_view=state_view,
        movement=movement,
        tip_handler=tip_handler,
        model_utils=model_utils,
        gantry_mover=gantry_mover,
        hardware_api=ot3_hardware_api,
        pipetting=pipetting,
    )

    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    error_id = "error-id"
    error_created_at = datetime(1234, 5, 6)

    decoy.when(
        state_view.geometry.convert_pick_up_tip_well_location(
            well_location=PickUpTipWellLocation(offset=WellOffset())
        )
    ).then_return(WellLocation(offset=WellOffset()))

    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(offset=WellOffset()),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_raise(StallOrCollisionDetectedError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_created_at)
    decoy.when(state_view.labware.get_definition(labware_id)).then_return(
        evotips_definition
    )

    result = await subject.execute(
        SealPipetteToTipParams(
            pipetteId=pipette_id, labwareId=labware_id, wellName=well_name
        )
    )

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=error_id, createdAt=error_created_at, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.CLEAR,
        ),
    )



"""Tests for the legacy Protocol API module core implementations."""

import pytest
from decoy import Decoy

from opentrons.drivers.types import ThermocyclerLidStatus
from opentrons.hardware_control import SyncHardwareAPI, SynchronousAdapter
from opentrons.hardware_control.modules import TemperatureStatus, Thermocycler
from opentrons.hardware_control.modules.types import (
    ThermocyclerModuleModel,
)
from opentrons.hardware_control.types import Axis
from opentrons.protocol_api.core.legacy.legacy_instrument_core import (
    LegacyInstrumentCore,
)
from opentrons.protocol_api.core.legacy.legacy_module_core import (
    LegacyThermocyclerCore,
    create_module_core,
)
from opentrons.protocol_api.core.legacy.legacy_protocol_core import (
    LegacyProtocolCore,
)
from opentrons.protocol_api.core.legacy.legacy_well_core import LegacyWellCore
from opentrons.protocol_api.core.legacy.module_geometry import (
    ThermocyclerGeometry,
)
from opentrons.protocol_api.labware import Labware
from opentrons.types import Location, Mount, Point

SyncThermocyclerHardware = SynchronousAdapter[Thermocycler]


@pytest.fixture
def mock_sync_hardware_api(decoy: Decoy) -> SyncHardwareAPI:
    """Get a mock sync hardware API."""
    return decoy.mock(cls=SyncHardwareAPI)


@pytest.fixture
def mock_protocol_core(
    decoy: Decoy, mock_sync_hardware_api: SyncHardwareAPI
) -> LegacyProtocolCore:
    """Get a mock protocol core."""
    mock_protocol_core = decoy.mock(cls=LegacyProtocolCore)
    decoy.when(mock_protocol_core.get_hardware()).then_return(mock_sync_hardware_api)
    return mock_protocol_core


@pytest.fixture
def mock_geometry(decoy: Decoy) -> ThermocyclerGeometry:
    """Get a mock thermocycler geometry."""
    return decoy.mock(cls=ThermocyclerGeometry)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SyncThermocyclerHardware:
    """Get a mock module hardware control interface."""
    return decoy.mock(name="SyncThermocyclerHardware")  # type: ignore[no-any-return]


@pytest.fixture
def mock_instrument_core(decoy: Decoy) -> LegacyInstrumentCore:
    """Get a mock instrument core."""
    return decoy.mock(cls=LegacyInstrumentCore)


@pytest.fixture
def mock_labware(decoy: Decoy) -> Labware:
    """Get a mock labware."""
    return decoy.mock(cls=Labware)


@pytest.fixture
def mock_trash_well_core(decoy: Decoy) -> LegacyWellCore:
    """Get a mock well implementation for the trash."""
    return decoy.mock(cls=LegacyWellCore)


@pytest.fixture
def subject(
    mock_geometry: ThermocyclerGeometry,
    mock_protocol_core: LegacyProtocolCore,
    mock_sync_module_hardware: SyncThermocyclerHardware,
) -> LegacyThermocyclerCore:
    """Get a legacy module implementation core with mocked out dependencies."""
    return LegacyThermocyclerCore(
        requested_model=ThermocyclerModuleModel.THERMOCYCLER_V1,
        protocol_core=mock_protocol_core,
        geometry=mock_geometry,
        sync_module_hardware=mock_sync_module_hardware,
    )


def test_create(
    decoy: Decoy,
    mock_geometry: ThermocyclerGeometry,
    mock_protocol_core: LegacyProtocolCore,
) -> None:
    """It should be able to create a thermocycler module core."""
    mock_module_hardware_api = decoy.mock(cls=Thermocycler)
    result = create_module_core(
        geometry=mock_geometry,
        protocol_core=mock_protocol_core,
        module_hardware_api=mock_module_hardware_api,
        requested_model=ThermocyclerModuleModel.THERMOCYCLER_V1,
    )

    assert isinstance(result, LegacyThermocyclerCore)


def test_get_lid_position(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the current lid position."""
    decoy.when(mock_sync_module_hardware.lid_status).then_return(
        ThermocyclerLidStatus.OPEN
    )
    result = subject.get_lid_position()
    assert result == ThermocyclerLidStatus.OPEN


def test_get_block_temperature_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the current block temperature status."""
    decoy.when(mock_sync_module_hardware.status).then_return(TemperatureStatus.IDLE)
    result = subject.get_block_temperature_status()
    assert result == TemperatureStatus.IDLE


def test_get_lid_temperature_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the current lid temperature status."""
    decoy.when(mock_sync_module_hardware.lid_temp_status).then_return(
        TemperatureStatus.IDLE
    )
    result = subject.get_lid_temperature_status()
    assert result == TemperatureStatus.IDLE


def test_get_block_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the current block temperature."""
    decoy.when(mock_sync_module_hardware.temperature).then_return(12.3)
    result = subject.get_block_temperature()
    assert result == 12.3


def test_get_block_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the target block temperature."""
    decoy.when(mock_sync_module_hardware.target).then_return(12.3)
    result = subject.get_block_target_temperature()
    assert result == 12.3


def test_get_lid_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the current lid temperature."""
    decoy.when(mock_sync_module_hardware.lid_temp).then_return(42.0)
    result = subject.get_lid_temperature()
    assert result == 42.0


def test_get_lid_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the current lid temperature."""
    decoy.when(mock_sync_module_hardware.lid_target).then_return(42.0)
    result = subject.get_lid_target_temperature()
    assert result == 42.0


def test_get_ramp_rate(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the ramp rate."""
    decoy.when(mock_sync_module_hardware.ramp_rate).then_return(1.23)
    result = subject.get_ramp_rate()
    assert result == 1.23


def test_get_hold_time(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the hold time."""
    decoy.when(mock_sync_module_hardware.hold_time).then_return(13.37)
    result = subject.get_hold_time()
    assert result == 13.37


def test_get_total_cycle_count(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the total cycle count."""
    decoy.when(mock_sync_module_hardware.total_cycle_count).then_return(321)
    result = subject.get_total_cycle_count()
    assert result == 321


def test_get_current_cycle_index(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the current cycle index."""
    decoy.when(mock_sync_module_hardware.current_cycle_index).then_return(123)
    result = subject.get_current_cycle_index()
    assert result == 123


def test_get_total_step_count(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the total step count."""
    decoy.when(mock_sync_module_hardware.total_step_count).then_return(1337)
    result = subject.get_total_step_count()
    assert result == 1337


def test_get_current_step_index(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should report the current step index."""
    decoy.when(mock_sync_module_hardware.current_step_index).then_return(42)
    result = subject.get_current_step_index()
    assert result == 42


def test_open_lid(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_protocol_core: LegacyProtocolCore,
    mock_instrument_core: LegacyInstrumentCore,
    mock_geometry: ThermocyclerGeometry,
    mock_trash_well_core: LegacyWellCore,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should open the lid with the hardware."""
    decoy.when(mock_protocol_core.get_loaded_instruments()).then_return(
        {Mount.RIGHT: mock_instrument_core}
    )
    decoy.when(mock_instrument_core.get_mount()).then_return(Mount.RIGHT)
    decoy.when(mock_sync_hardware_api.current_position(Mount.RIGHT)).then_return(
        {Axis.A: 4}
    )
    decoy.when(mock_protocol_core.fixed_trash.get_well_core("A1")).then_return(
        mock_trash_well_core
    )
    decoy.when(mock_trash_well_core.get_top(z_offset=0)).then_return(
        Point(x=1, y=2, z=3)
    )

    decoy.when(mock_sync_module_hardware.open()).then_return("open")

    result = subject.open_lid()

    decoy.verify(
        mock_sync_hardware_api.retract(Mount.RIGHT),
        mock_instrument_core.move_to(
            Location(Point(x=1, y=2, z=4), None),
            well_core=None,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        decoy.prop(mock_geometry.lid_status).set(ThermocyclerLidStatus.OPEN),
    )
    assert result == ThermocyclerLidStatus.OPEN


def test_close_lid(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_protocol_core: LegacyProtocolCore,
    mock_instrument_core: LegacyInstrumentCore,
    mock_geometry: ThermocyclerGeometry,
    mock_trash_well_core: LegacyWellCore,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should close the lid with the hardware."""
    decoy.when(mock_protocol_core.get_loaded_instruments()).then_return(
        {Mount.LEFT: mock_instrument_core}
    )
    decoy.when(mock_instrument_core.get_mount()).then_return(Mount.LEFT)
    decoy.when(mock_sync_hardware_api.current_position(Mount.LEFT)).then_return(
        {Axis.Z: 4}
    )
    decoy.when(mock_protocol_core.fixed_trash.get_well_core("A1")).then_return(
        mock_trash_well_core
    )
    decoy.when(mock_trash_well_core.get_top(z_offset=0)).then_return(
        Point(x=1, y=2, z=3)
    )

    decoy.when(mock_sync_module_hardware.close()).then_return("closed")

    result = subject.close_lid()

    decoy.verify(
        mock_sync_hardware_api.retract(Mount.LEFT),
        mock_instrument_core.move_to(
            location=Location(Point(x=1, y=2, z=4), None),
            well_core=None,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        decoy.prop(mock_geometry.lid_status).set(ThermocyclerLidStatus.CLOSED),
    )
    assert result == ThermocyclerLidStatus.CLOSED


def test_set_target_block_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should set the block temperature with the hardware."""
    subject.set_target_block_temperature(
        celsius=42.0,
        hold_time_seconds=1.2,
        block_max_volume=3.4,
        ramp_rate=None,
    )

    decoy.verify(
        mock_sync_module_hardware.set_target_block_temperature(
            celsius=42.0,
            hold_time_seconds=1.2,
            volume=3.4,
            ramp_rate=None,
        ),
        times=1,
    )


def test_wait_for_block_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should wait for the block temperature with the hardware."""
    subject.wait_for_block_temperature()

    decoy.verify(mock_sync_module_hardware.wait_for_block_target(), times=1)


def test_set_target_lid_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should set the lid temperature with the hardware."""
    subject.set_target_lid_temperature(celsius=42.0)

    decoy.verify(mock_sync_module_hardware.set_target_lid_temperature(celsius=42.0))


def test_wait_for_lid_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should wait for the lid temperature with the hardware."""
    subject.wait_for_lid_temperature()

    decoy.verify(mock_sync_module_hardware.wait_for_lid_target(), times=1)


def test_execute_profile(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should execute a valid profile with the hardware."""
    subject.execute_profile(
        steps=[{"temperature": 42.0, "hold_time_seconds": 45.6}],
        repetitions=12,
        block_max_volume=34.5,
    )

    decoy.verify(
        mock_sync_module_hardware.cycle_temperatures(
            steps=[{"temperature": 42.0, "hold_time_seconds": 45.6}],
            repetitions=12,
            volume=34.5,
        )
    )


def test_deactivate_lid(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should turn off the heated lid with the hardware."""
    subject.deactivate_lid()

    decoy.verify(mock_sync_module_hardware.deactivate_lid(), times=1)


def test_deactivate_block(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should turn off the well block temperature controller with the hardware."""
    subject.deactivate_block()

    decoy.verify(mock_sync_module_hardware.deactivate_block(), times=1)


def test_deactivate(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: LegacyThermocyclerCore,
) -> None:
    """It should turn off the well block and lid temperature with the hardware."""
    subject.deactivate()

    decoy.verify(mock_sync_module_hardware.deactivate(), times=1)
